from sqlalchemy import String, Column, PickleType
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


# log url {nsrId}/vnfrecord/{vnfrName}/hostname/{hostname}
class Nsr(Base):
    __tablename__ = "ns_records"

    id = Column(String(250), primary_key=True)
    username = Column(String(250), nullable=False)
    status = Column(String(250), nullable=False)
    # {vnfrName:hostname}
    vnf_log_url = Column(PickleType)



import json
import logging
import logging.config
import os

from sdk.softfire.utils import get_config, TESTBED_MAPPING

from eu.softfire.nfv.utils.static_config import CONFIG_FILE_PATH


def get_logger(name):
    logging.config.fileConfig(CONFIG_FILE_PATH, disable_existing_loggers=False)
    if name.startswith("eu.softfire.nfv"):
        return logging.getLogger(name)
    return logging.getLogger("eu.softfire.nfv.%s" % name)


logger = get_logger(__name__)


def get_available_nsds():
    with open(get_config('system', 'available-nsds-file-path', CONFIG_FILE_PATH), 'r') as f:
        return json.loads(f.read())


def get_openstack_credentials():
    openstack_credential_file_path = get_config('system', 'openstack-credentials-file', CONFIG_FILE_PATH)
    # logger.debug("Openstack cred file is: %s" % openstack_credential_file_path)
    if os.path.exists(openstack_credential_file_path):
        with open(openstack_credential_file_path, "r") as f:
            return json.loads(f.read())
    else:
        raise FileNotFoundError("Openstack credentials file not found")


def get_testbed_name_from_id(testbed_id):
    for k, v in TESTBED_MAPPING.items():
        if v == testbed_id:
            return k



"""
Contains the static configuration of the system
"""
import os

if os.path.exists('/etc/softfire/nfv-manager.ini'):
    CONFIG_FILE_PATH = '/etc/softfire/nfv-manager.ini'
else:
    CONFIG_FILE_PATH = 'etc/nfv-manager.ini'



class _BaseException(Exception):
    def __init__(self, message=None) -> None:
        super().__init__()
        self.message = message


class OpenstackClientError(_BaseException):
    pass


class NfvResourceValidationError(_BaseException):
    pass


class NfvResourceDeleteException(_BaseException):
    pass


class NfvResourceProvisioningException(_BaseException):
    pass


class MissingFileException(NfvResourceProvisioningException):
    pass


class NfvManagerNotFoundException(_BaseException):
    pass



import unittest

from org.openbaton.cli.agents.agents import OpenBatonAgentFactory

from eu.softfire.nfv.utils import get_config


class MyTestCase(unittest.TestCase):
    def test_openbaton(self):
        config = get_config()
        agent = OpenBatonAgentFactory(nfvo_ip=config.get("nfvo", "ip"),
                                      nfvo_port=config.get("nfvo", "port"),
                                      https=config.getboolean("nfvo", "https"),
                                      version=1,
                                      username=config.get("nfvo", "username"),
                                      password=config.get("nfvo", "password"),
                                      project_id=None)
        for pj in agent.get_project_agent().find():
            print(pj)

if __name__ == '__main__':
    unittest.main()



import json
# from utils import os_utils_opnfv
import os
import time
import traceback
from os import listdir
from os.path import isfile, join
from threading import Thread

import yaml
from org.openbaton.cli.errors.errors import NfvoException
from org.openbaton.cli.openbaton import LIST_PRINT_KEY
from sdk.softfire.grpc import messages_pb2
from sdk.softfire.manager import AbstractManager
from sdk.softfire.utils import TESTBED_MAPPING
from sqlalchemy.orm.exc import NoResultFound

from eu.softfire.nfv.db.entities import Nsr
from eu.softfire.nfv.db.repositories import find, delete, save
from eu.softfire.nfv.utils import os_utils
from eu.softfire.nfv.utils.exceptions import NfvResourceValidationError, NfvResourceDeleteException, \
    MissingFileException, NfvManagerNotFoundException
from eu.softfire.nfv.utils.ob_utils import OBClient
from eu.softfire.nfv.utils.os_utils import create_os_project
from eu.softfire.nfv.utils.static_config import CONFIG_FILE_PATH
from eu.softfire.nfv.utils.utils import get_available_nsds, get_logger, get_config

logger = get_logger(__name__)

AVAILABLE_AGENTS = LIST_PRINT_KEY.keys()

CARDINALITY = {
    'open5gcore': 1,
}


class UpdateStatusThread(Thread):
    def __init__(self, nfv_manager):
        Thread.__init__(self)
        self.stopped = False
        self.nfv_manager = nfv_manager

    def run(self):
        while not self.stopped:
            time.sleep(int(self.nfv_manager.get_config_value('system', 'update-delay', '10')))
            if not self.stopped:
                try:
                    self.nfv_manager.send_update()
                except Exception as e:
                    traceback.print_exc()
                    if hasattr(e, 'args'):
                        logger.error("got error while updating resources: %s " % e)
                    else:
                        logger.error("got unknown error while updating resources")

    def stop(self):
        self.stopped = True


def get_nsrs_to_check():
    return find(Nsr)


def add_nsr_to_check(username, nsr):
    try:
        nsr_exists = find(Nsr, _id=nsr.get('id'))
        if nsr_exists:
            delete(nsr_exists)
    except NoResultFound:
        pass

    nsr_to_save = Nsr()
    nsr_to_save.username = username
    nsr_to_save.id = nsr.get('id')
    nsr_to_save.status = nsr.get('status')
    nsr_to_save.vnf_log_url = {}

    for vnfr in nsr.get('vnfr'):
        for vdu in vnfr.get('vdu'):
            for vnfc_instance in vdu.get('vnfc_instance'):
                if not nsr_to_save.vnf_log_url.get(vnfr.get('name')):
                    nsr_to_save.vnf_log_url[vnfr.get('name')] = ""
                nsr_to_save.vnf_log_url[vnfr.get('name')] += "%s;" % vnfc_instance.get('hostname')
    save(nsr_to_save, Nsr)


def remove_nsr_to_check(nsr_id):
    try:
        delete(find(Nsr, _id=nsr_id))
    except NoResultFound:
        pass


def _update_nsr(nsr):
    logger.debug("Checking resources of user %s, nsr id %s" % (nsr.username, nsr.id))
    ob_client = OBClient(nsr.username)
    # check if the OBClient has a project ID. if not, something went wrong and the nsr is ignored for now.
    if ob_client.project_id is None:
        logger.error('The OBClient for user {} has no project ID. This should never happen. Does the user still exist '
                     'in Open Baton?'.format(nsr.username))
        raise NfvManagerNotFoundException(
            'The OBClient for user {} has no project ID. This should never happen. Does the user still exist in Open '
            'Baton?'.format(
                nsr.username))
    try:
        nsr_new = ob_client.get_nsr(nsr.id)
    except Exception as e:
        logger.error(
            'Exception while fetching the NSR with ID {} of user {}. Removing it.'.format(nsr.id, nsr.username))
        remove_nsr_to_check(nsr.id)
        raise NfvManagerNotFoundException(
            'Exception while fetching the NSR with ID {} of user {}. Does it really exist?'.format(nsr.id,
                                                                                                   nsr.username))
    try:
        nsr_new_dict = json.loads(nsr_new)
    except:
        logger.error('Not able to parse nsr to dictionary: {}'.format(nsr_new))
        return None
    if 'error' in nsr_new_dict:
        logger.error('Exception while updating the NSR with ID {} of user {}: {}'.format(nsr.id, nsr.username, nsr_new_dict.get('error')))
        return None
    add_nsr_to_check(nsr.username, nsr_new_dict)
    logger.debug("Status is: %s" % nsr_new_dict.get('status'))
    return nsr_new


def try_delete_vnfd(vnfd_id, ob_client):
    try:
        ob_client.delete_vnfd(vnfd_id)
    except:
        raise NfvResourceDeleteException('Not able to delete VNFD with id: %s' % vnfd_id)


def try_delete_nsd(nsd_id, ob_client):
    try:
        ob_client.delete_nsd(nsd_id)
    except:
        raise NfvResourceDeleteException('Not able to delete NSD with id: %s' % nsd_id)


def try_delete_nsr(nsr, ob_client):
    try:
        ob_client.delete_nsr(nsr.get('id'))
    except:
        raise NfvResourceDeleteException('Not able to delete NSR with id: %s' % nsr.get('id'))
    timer = 500
    time.sleep(5)
    while True:
        try:
            nsr = json.loads(ob_client.get_nsr(nsr.get('id')))
            if timer == 0:
                raise NfvResourceDeleteException('Not able to delete NSR with id: %s' % nsr.get('id'))
        except NfvoException:
            return
        timer -= 1
        time.sleep(2)


def remove_all(ob_client, force=False):
    if force or get_config('system', 'delete-all', CONFIG_FILE_PATH, 'false').lower() == 'true':
        logger.debug("removing everything!")
        for _nsr in ob_client.list_nsrs():
            ob_client.delete_nsr(_nsr.get('id'))
        for _nsd in ob_client.list_nsds():
            ob_client.delete_nsd(_nsd.get('id'))
        for _vim_instance in ob_client.list_vim_instances():
            ob_client.delete_vim_instance(_vim_instance.get('id'))


class NfvManager(AbstractManager):
    def __init__(self, config_file_path):
        super().__init__(config_file_path)
        with open(self.get_config_value('system', 'softfire-public-key'), "r") as sosftfire_ssh_pub_key:
            self.softfire_pub_key = sosftfire_ssh_pub_key.read().strip()

    def validate_resources(self, user_info=None, payload=None) -> None:

        request_dict = yaml.load(payload)
        logger.info("Validating %s " % request_dict)

        resource_id = request_dict.get("properties").get('resource_id')
        available_nsds = get_available_nsds()
        if resource_id not in available_nsds.keys():
            temp_csar_location = self.get_config_value('system', 'temp-csar-location',
                                                       '/etc/softfire/experiment-nsd-csar').rstrip('/')
            nsd_location = '{}/{}/{}.csar'.format(temp_csar_location, user_info.name, resource_id)
            file_name = request_dict.get("properties").get('file_name')
            logger.debug("Checking if nsd_location exists: %s and if there is filename: %s" % (
                os.path.exists(nsd_location), file_name))
            if not os.path.exists(nsd_location) and not file_name:
                raise NfvResourceValidationError(
                    message="Resource id %s not in the available ones %s and no CSAR file provided" % (
                        resource_id, list(available_nsds.keys())))

        else:
            testbeds = request_dict.get("properties").get("testbeds")
            for vnf_type in testbeds.keys():
                if vnf_type not in available_nsds.get(resource_id).get("vnf_types") and vnf_type.upper() != "ANY":
                    raise NfvResourceValidationError(
                        message="Testbeds properties must be a dict containing the vnf type of the NS chosen or ANY, "
                                "%s not included in the possibilities %s" % (
                                    vnf_type, available_nsds.get("resource_id").get("vnf_types")))
        pass

    def refresh_resources(self, user_info):
        """
            List all available images for this tenant

            :param user_info:
            :return: the list of ResourceMetadata
             :rtype list
            """
        result = []
        ob_client = OBClient(user_info.name)
        images, networks, flavours = ob_client.list_images_network_flavors()

        for image in images:
            testbed = image.get('testbed')
            resource_id = image.get('name')
            result.append(messages_pb2.ResourceMetadata(resource_id=resource_id,
                                                        description='',
                                                        cardinality=-1,
                                                        node_type='NfvImage',
                                                        testbed=TESTBED_MAPPING.get(testbed)))
        for net in networks:
            testbed = net.get('testbed')
            resource_id = net.get('name')
            result.append(messages_pb2.ResourceMetadata(resource_id=resource_id,
                                                        description='',
                                                        cardinality=-1,
                                                        node_type='NfvNetwork',
                                                        testbed=TESTBED_MAPPING.get(testbed)))
        for flavour in flavours:
            testbed = flavour.get('testbed')
            resource_id = flavour.get('name')
            result.append(messages_pb2.ResourceMetadata(resource_id=resource_id,
                                                        description='',
                                                        cardinality=-1,
                                                        node_type='NfvFlavor',
                                                        testbed=TESTBED_MAPPING.get(testbed)))
        return result

    def provide_resources(self, user_info, payload=None):
        """
            Deploy the selected resources. Payload looks like:
            {
                'properties': {
                    'nsd_name': 'my_nsd',
                    'file_name': 'Files/my_nsd.csar',
                    'resource_id': 'open5gcore',
                    'testbeds': {
                        'ANY':
                        'fokus'
                    }
                },
                'type': 'NfvResource'
            }

            :param payload: the resources to be deployed
             :type payload: dict
            :param user_info: the user info requesting
            :return: the nsr deployed
             :rtype: ProvideResourceResponse
            """
        ob_client = OBClient(user_info.name)
        logger.debug("Payload is \n%s" % payload)
        resource_dict = json.loads(payload)
        logger.debug("Received %s " % resource_dict)
        ssh_pub_key = resource_dict.get("properties").get('ssh_pub_key')
        resource_id = resource_dict.get("properties").get("resource_id")
        monitoring_ip = resource_dict.get("properties").get("floatingIp")
        if not monitoring_ip:
            monitoring_ip = ""
        file_name = resource_dict.get("properties").get("file_name")
        nsd_name = resource_dict.get("properties").get("nsd_name")

        ob_client.import_key(self.softfire_pub_key, 'softfire-key')
        nsr_keys_to_use = ['softfire-key']
        if ssh_pub_key:
            logger.debug("creating user-key called: %s" % nsd_name)
            ob_client.import_key(ssh_pub_key.strip(), nsd_name)
            nsr_keys_to_use.append(nsd_name)

        # temp_csar_location = "{}/{}".format(
        #     self.get_config_value("system", "temp-csar-location", '/etc/softfire/experiment-nsd-csar').rstrip('/'),
        #     resource_id)
        available_nsds = get_available_nsds()
        nsd_chosen = available_nsds.get(resource_id)
        packages_location = "%s/%s" % (
            self.get_config_value("system", "packages-location", '/etc/softfire/packages').rstrip('/'), resource_id)
        vnfds = []
        testbeds = resource_dict.get("properties").get("testbeds")

        logger.debug("Checking if nsd_chosen is not none: %s" % nsd_chosen)
        logger.debug("and if path %s exists: %s" % (packages_location, os.path.exists(packages_location)))

        if nsd_chosen and os.path.exists(packages_location):
            for package in [f for f in listdir(packages_location) if isfile(join(packages_location, f))]:
                vnfd = ob_client.upload_package(join(packages_location, package), package.split('.')[0])
                vnfds.append({
                    'id': vnfd.get('id')
                })

            virtual_links = [{"name": "softfire-internal"}]
            nsd = {
                "name": nsd_name,
                "version": "softfire_version",
                "vendor": user_info.name,
                "vnfd": vnfds,
                "vld": virtual_links
            }

            logger.debug("Uploading NSD: %s" % nsd)

            nsd = ob_client.create_nsd(nsd)

            logger.debug("Created NSD: %s" % nsd)
            vdu_vim_instances = {}

            if "ANY" in testbeds.keys():
                for vdu_name in nsd_chosen.get("vnf_types"):
                    vdu_vim_instances[vdu_name] = ["vim-instance-%s" % vim_name for vim_name in testbeds.values()]
            else:
                for vdu_name in nsd_chosen.get("vnf_types"):
                    vdu_vim_instances[vdu_name] = ["vim-instance-%s" % testbeds.get(vdu_name)]

            body = json.dumps({
                "vduVimInstances": vdu_vim_instances,
                "keys": nsr_keys_to_use,
                "monitoringIp": monitoring_ip
            })
            logger.debug("Body is %s" % body)
            try:
                nsr = ob_client.create_nsr(nsd.get('id'), body=body)
            except Exception as e:
                logger.error('Exception while deploying NSR from NSD: {}'.format(e))
                logger.debug('Delete NSD {}'.format(nsd.get('id')))
                try:
                    ob_client.delete_nsd(nsd.get('id'))
                except Exception as e2:
                    logger.error('Could not remove NSD {}: {}'.format(nsd.get('id'), e2))
                raise e

            add_nsr_to_check(user_info.name, nsr)

        else:
            # nsd resource was added by the user and is not available for everyone
            temp_csar_location = self.get_config_value("system", "temp-csar-location",
                                                       "/etc/softfire/experiment-nsd-csar")
            csar_nsd_file_path = "{}/{}/{}".format(
                temp_csar_location.rstrip('/'), user_info.name, file_name[6:])
            if os.path.exists(csar_nsd_file_path):
                nsd = ob_client.create_nsd_from_csar(csar_nsd_file_path)
                logger.debug("Created NSD: %s" % nsd.get('name'))
            else:
                raise MissingFileException("File %s was not found" % csar_nsd_file_path)
            vdu_vim_instances = {}

            if "ANY" in testbeds.keys():
                for vnfd in nsd.get('vnfd'):
                    for vdu in vnfd.get('vdu'):
                        vdu_vim_instances[vdu.get('name')] = ["vim-instance-%s" % vim_name for vim_name in
                                                              testbeds.values()]
            else:
                logger.debug("Adding testbed mapping vdu : %s" % testbeds)
                for vdu_name, testbed in testbeds.items():
                    vdu_vim_instances[vdu_name] = ["vim-instance-%s" % testbed]
            body = json.dumps({
                "vduVimInstances": vdu_vim_instances,
                "keys": nsr_keys_to_use,
                "monitoringIp": monitoring_ip
            })
            logger.debug("Deploy NSR with body: %s" % body)

            if nsd:
                try:
                    nsr = ob_client.create_nsr(nsd.get('id'), body=body)
                except Exception as e:
                    logger.error('Exception while deploying NSR from NSD {}: {}'.format(nsd.get('id'), e))
                    logger.debug('Delete NSD {}'.format(nsd.get('id')))
                    try:
                        ob_client.delete_nsd(nsd.get('id'))
                    except Exception as e2:
                        logger.error('Could not remove NSD {}: {}'.format(nsd.get('id'), e2))
                    raise e
                add_nsr_to_check(user_info.name, nsr)



        if isinstance(nsr, dict):
            nsr = json.dumps(nsr)

        logger.info("Deployed resource: %s" % json.loads(nsr).get('name'))
        return [nsr]

    def create_user(self, user_info):
        """
            Create project in Open Stack and upload the new vim to Open Baton

            :param user_info:
            :return: the new user info updated
             :rtype: UserInfo

            """
        username = user_info.name
        password = user_info.password
        os_tenants = create_os_project(username=username, password=password, tenant_name=username)
        ob_client = OBClient()
        project = {
            'name': username,
            'description': 'the project for user %s' % username
        }
        project = ob_client.create_project(project)
        user = {
            'username': username,
            'password': password,
            'enabled':  True,
            'email':    "{}@softfire.local".format(username),
            'roles':    [
                {
                    'role': 'USER',
                    'project': project.get('name')
                }
            ]
        }
        logger.debug("Create openbaton project %s" % project)
        ob_client = OBClient(project.get('name'))
        user = ob_client.create_user(user)
        logger.debug("Create openbaton user %s" % user)

        user_info.ob_project_id = project.get('id')
        # user_info.testbed_tenants = {}

        testbed_tenants = {}
        if os_tenants:
            for testbed_name, v in os_tenants.items():
                tenant_id = v.get('tenant_id')
                vim_instance = v.get('vim_instance')
                try:
                    vi = ob_client.create_vim_instance(vim_instance)
                    logger.debug("created vim instance with id: %s" % vi.get('id'))
                except NfvoException:
                    logger.warning("Not able to upload vim %s" % testbed_name)
                testbed_tenants[TESTBED_MAPPING[testbed_name]] = tenant_id

        for k, v in testbed_tenants.items():
            user_info.testbed_tenants[k] = v
        logger.debug("Updated user_info %s" % user_info)

        return user_info

    def list_resources(self, user_info=None, payload=None):
        """
            list all available resources

            :param payload: Not used
            :param user_info: the user info requesting, if None only the shared resources will be returned 
            :return: list of ResourceMetadata
            """
        result = []

        if user_info and user_info.name:
            ob_client = OBClient(user_info.name)

            for nsd in ob_client.list_nsds():
                result.append(messages_pb2.ResourceMetadata(nsd.name,
                                                            nsd.get('description') or get_available_nsds[
                                                                nsd.name.lower()].get(
                                                                'description'),
                                                            CARDINALITY[nsd.name.lower()]))

                result.extend(self.refresh_resources(user_info))

        for k, v in get_available_nsds().items():
            testbed = v.get('testbed')
            node_type = v.get('node_type')
            cardinality = int(v.get('cardinality'))
            description = v.get('description')
            resource_id = k
            result.append(messages_pb2.ResourceMetadata(resource_id=resource_id,
                                                        description=description,
                                                        cardinality=cardinality,
                                                        node_type=node_type,
                                                        testbed=TESTBED_MAPPING.get(testbed)))

        return result

    def release_resources(self, user_info, payload=None):
        """
           Delete the NSR from openbaton based on user_info and the nsr
           :param payload: the NSR itself
           :type payload: dict
           :param user_info:
            :type user_info: UserInfo
           :return: None
           """
        ob_client = OBClient(user_info.name)

        logger.info('Deleting resources for user: {}'.format(user_info.name))
        logger.debug('Received this payload: {}'.format(payload))
        nsr = None
        try:
            nsr = json.loads(payload)
            assert type(nsr) == dict
        except:
            logger.warning('Could not parse release resource payload to JSON: {}'.format(payload))
            traceback.print_exc()
            if nsr:
                try:
                    remove_nsr_to_check(nsr.get('id'))
                except:
                    pass
            return
        if nsr.get('type') == 'NfvResource' and nsr.get('properties') is not None:
            logger.debug('The payload does not seem to be an NSR so the resource was probably not yet deployed and '
                         'nothing has to be removed from Open Baton.')
            return
        nsd_id = nsr.get('descriptor_reference')
        try:
            nsd = json.loads(ob_client.get_nsd(nsd_id))
        except NfvoException:
            traceback.print_exc()
            return
        vnfd_ids = []
        for vnfd in nsd.get('vnfd'):
            vnfd_ids.append(vnfd.get('id'))

        try:
            try_delete_nsr(nsr, ob_client)

            try_delete_nsd(nsd_id, ob_client)
            # TODO to be added if cascade is not enabled
            # for vnfd_id in vnfd_ids:
            #     self.try_delete_vnfd(vnfd_id, ob_client)
        except NfvResourceDeleteException as e:
            traceback.print_exc()
            logger.error("...ignoring...")

        remove_nsr_to_check(nsr.get('id'))
        logger.info("Removed resource %s" % nsr.get('name'))

        # remove_all(ob_client)

    def _update_status(self) -> dict:
        result = {}
        for nsrs in get_nsrs_to_check():
            if isinstance(nsrs, list):
                for nsr in nsrs:
                    if not result.get(nsr.username):
                        result[nsr.username] = []
                    if nsr.status.lower() not in ['active']:
                        try:
                            nsr_new = _update_nsr(nsr)
                        except NfvManagerNotFoundException:
                            remove_nsr_to_check(nsr.id)
                            continue
                        if nsr_new is not None:
                            result[nsr.username].append(nsr_new)
            else:
                if not result.get(nsrs.username):
                    result[nsrs.username] = []
                if nsrs.status.lower() not in ['active']:
                    try:
                        nsr_new = _update_nsr(nsrs)
                    except NfvManagerNotFoundException:
                        remove_nsr_to_check(nsrs.id)
                        continue
                    if nsr_new is not None:
                        result[nsrs.username].append(nsr_new)

        return result

    def delete_user(self, user_info):
        logger.debug("Removing user %s" % user_info)
        username = user_info.name
        ob_client = OBClient(username)
        try:
            remove_all(ob_client, True)
        except:
            pass
        time.sleep(5)
        try:
            ob_client.delete_user(username=username)
        except:
            pass
        try:
            ob_client.delete_project(ob_project_id=user_info.ob_project_id)
        except:
            pass
        time.sleep(2)
        try:
            os_utils.delete_tenant_and_user(username=username, testbed_tenants=user_info.testbed_tenants)
        except:
            pass



import logging
import traceback

import keystoneclient
import neutronclient
from glanceclient import Client as Glance
from keystoneauth1 import session
from keystoneauth1.exceptions.http import Conflict
from keystoneauth1.identity import v2, v3
from neutronclient.common.exceptions import IpAddressGenerationFailureClient
from neutronclient.v2_0.client import Client as Neutron
from novaclient.client import Client as Nova

from eu.softfire.nfv.utils.exceptions import OpenstackClientError
from eu.softfire.nfv.utils.utils import get_logger, get_config, get_openstack_credentials, get_testbed_name_from_id

logger = get_logger(__name__)

NETWORKS = ["mgmt", "net_a", "net_b", "net_c", "net_d", "private", "softfire-internal"]
sec_group_name = 'ob_sec_group'


class OSClient(object):
    def __init__(self, testbed_name, testbed, tenant_name=None, project_id=None):
        self.testbed_name = testbed_name
        self.tenant_name = None
        self.project_id = None
        self.testbed = testbed
        self.project_domain_name = self.testbed.get('project_domain_name') or 'Default'
        self.user_domain_id = self.testbed.get('user_domain_id') or 'Default'
        self.api_version = self.testbed.get('api_version')
        self.username = self.testbed.get('username')
        self.password = self.testbed.get('password')
        self.auth_url = self.testbed.get("auth_url")
        if self.auth_url.endswith('/'):
            self.auth_url = self.auth_url[:-1]
        self.admin_tenant_name = self.testbed.get("admin_tenant_name")
        self.admin_project_id = self.testbed.get("admin_project_id")
        if not self.admin_tenant_name and not self.admin_project_id:
            raise OpenstackClientError("Missing both adimn project id and admin tenant name")
        if self.api_version == 2 and not self.admin_tenant_name:
            raise OpenstackClientError("Missing tenant name required if using v2")
        if self.api_version == 3 and not self.admin_project_id:
            raise OpenstackClientError("Missing project id required if using v3")

        self.neutron = None
        self.nova = None
        self.glance = None
        self.keypair = None
        self.sec_group = None
        self.os_tenant_id = None

        logger.debug("Log level is: %s and DEBUG is %s" % (logger.getEffectiveLevel(), logging.DEBUG))
        if logger.getEffectiveLevel() == logging.DEBUG:
            logging.basicConfig(level=logging.DEBUG)

        if not tenant_name and not project_id:

            self.keystone = self._create_keystone_client()
            logger.debug("Created Keystone client %s" % self.keystone)
        else:
            self.tenant_name = tenant_name
            self.project_id = project_id

            if self.api_version == 2 and not self.tenant_name:
                raise OpenstackClientError("Missing tenant name required if using v2")
            if self.api_version == 3 and not self.project_id:
                raise OpenstackClientError("Missing project id required if using v3")

            logger.debug("Creating keystone client")
            if self.api_version == 3:
                self.keystone = self._create_keystone_client(project_id)
                self.os_tenant_id = project_id
            else:
                self.keystone = self._create_keystone_client(tenant_name)
                self.os_tenant_id = self.project_id = self._get_tenant_id_from_name(tenant_name)

            logger.debug("Created Keystone client %s" % self.keystone)
            self.set_nova(self.os_tenant_id)
            self.set_neutron(self.os_tenant_id)
            self.set_glance(self.os_tenant_id)

    def _create_keystone_client(self, project_id=None):
        if self.api_version == 3:
            return keystoneclient.v3.client.Client(session=self._get_session(project_id))
        elif self.api_version == 2:
            if not project_id:
                project_id = self.tenant_name or self.admin_tenant_name
            return keystoneclient.v2_0.client.Client(username=self.username,
                                                     password=self.password,
                                                     tenant_name=project_id,
                                                     auth_url=self.auth_url)

    def set_nova(self, os_tenant_id):
        self.nova = Nova('2.1', session=self._get_session(os_tenant_id))

    def _get_session(self, tenant_id=None):
        if self.api_version == 2:
            tenant_name = self.tenant_name or self.admin_tenant_name
            auth = v2.Password(auth_url=self.auth_url,
                               username=self.username,
                               password=self.password,
                               tenant_name=tenant_name)
        elif self.api_version == 3:
            p_id = tenant_id or self.project_id or self.admin_project_id
            auth = v3.Password(auth_url=self.auth_url,
                               username=self.username,
                               password=self.password,
                               project_id=p_id,
                               project_domain_name=self.project_domain_name,
                               user_domain_id=self.user_domain_id)
        else:
            msg = "Wrong api version: %s" % self.api_version
            logger.error(msg)
            raise OpenstackClientError(msg)
        return session.Session(auth=auth)

    def set_neutron(self, os_tenant_id):
        # self.os_tenant_id = os_tenant_id
        if not self.neutron:
            self.neutron = Neutron(session=self._get_session(os_tenant_id))

    def get_user(self, username=None):
        users = self.list_users()
        if username:
            un = username
        else:
            un = self.username
        for user in users:
            if user.name == un:
                return user

    def get_role(self, role_to_find):
        roles = self.list_roles()
        for role in roles:
            if role.name == role_to_find:
                return role

    def list_roles(self):
        return self.keystone.roles.list()

    def list_tenants(self):
        if self.api_version == 3:
            return self.keystone.projects.list()
        else:
            return self.keystone.tenants.list()

    def create_tenant(self, tenant_name, description):
        self.tenant_name = tenant_name
        if self.api_version == 2:
            return self.keystone.tenants.create(tenant_name=tenant_name, description=description)
        else:
            return self.keystone.projects.create(name=tenant_name, description=description,
                                                 domain=self.user_domain_id)

    def add_user_role(self, user, role, tenant):
        if self.api_version == 2:
            try:
                return self.keystone.roles.add_user_role(user=user, role=role, tenant=tenant)
            except Conflict as c:
                if c.http_status == 409:  # role already assigned to user
                    return
                raise c
        else:
            return self.keystone.roles.grant(user=user, role=role, project=tenant)

    def import_keypair(self, os_tenant_id=None):
        if not self.nova and not os_tenant_id:
            raise OpenstackClientError("Both os_tenant_id and nova obj are None")
        if not self.nova:
            self.set_nova(os_tenant_id=os_tenant_id)
        keypair_name = "softfire-key"
        self.keypair = keypair_name
        for keypair in self.list_keypairs(os_tenant_id):
            if keypair.name == keypair_name:
                return keypair
        with open(get_config().get('system', 'softfire-public-key'), "r") as sosftfire_ssh_pub_key:
            kargs = {"name": keypair_name,
                     "public_key": sosftfire_ssh_pub_key.read()}
            return self.nova.keypairs.create(**kargs)

    def get_ext_net(self, ext_net_name='softfire-network'):
        return [ext_net for ext_net in self.neutron.list_networks()['networks'] if
                ext_net['router:external'] and ext_net['name'] == ext_net_name][0]

    def allocate_floating_ips(self, fip_num=0, ext_net='softfire-network'):
        body = {
            "floatingip": {
                "floating_network_id": ext_net['id']
            }
        }
        for i in range(fip_num):
            try:
                self.neutron.create_floatingip(body=body)
            except IpAddressGenerationFailureClient as e:
                logger.error("Not able to allocate floatingips :(")
                raise OpenstackClientError("Not able to allocate floatingips :(")

    def create_networks_and_subnets(self, ext_net, router_name='ob_router'):
        networks = []
        subnets = []
        ports = []
        router_id = None
        exist_net = [network for network in self.neutron.list_networks()['networks']]
        exist_net_names = [network['name'] for network in exist_net]
        net_name_to_create = [net for net in NETWORKS if net not in exist_net_names]
        networks.extend(network for network in exist_net if network['name'] in NETWORKS)
        index = 1
        for net in net_name_to_create:
            kwargs = {'network': {
                'name': net,
                'shared': False,
                'admin_state_up': True
            }}
            logger.debug("Creating net %s" % net)
            network_ = self.neutron.create_network(body=kwargs)['network']
            networks.append(network_)
            kwargs = {
                'subnets': [
                    {
                        'name': "subnet_%s" % net,
                        'cidr': "192.%s.%s.0/24" % ((get_username_hash(self.username) % 254) + 1, index),
                        'gateway_ip': '192.%s.%s.1' % ((get_username_hash(self.username) % 254) + 1, index),
                        'ip_version': '4',
                        'enable_dhcp': True,
                        'dns_nameservers': ['8.8.8.8'],
                        'network_id': network_['id']
                    }
                ]
            }
            logger.debug("Creating subnet subnet_%s" % net)
            subnet = self.neutron.create_subnet(body=kwargs)
            subnets.append(subnet)

            router = self.get_router_from_name(router_name, ext_net)
            router_id = router['router']['id']

            body_value = {
                'subnet_id': subnet['subnets'][0]['id'],
            }
            try:
                ports.append(self.neutron.add_interface_router(router=router_id, body=body_value))
            except Exception as e:
                pass
            index += 1

        return networks, subnets, router_id

    def get_router_from_name(self, router_name, ext_net):
        for router in self.neutron.list_routers()['routers']:
            if router['name'] == router_name:
                return self.neutron.show_router(router['id'])
        request = {'router': {'name': router_name, 'admin_state_up': True}}
        router = self.neutron.create_router(request)
        body_value = {"network_id": ext_net['id']}
        self.neutron.add_gateway_router(router=router['router']['id'], body=body_value)
        return router

    def create_rule(self, sec_group, protocol):
        body = {"security_group_rule": {
            "direction": "ingress",
            "port_range_min": "1",
            "port_range_max": "65535",
            # "name": sec_group['security_group']['name'],
            "security_group_id": sec_group['security_group']['id'],
            "remote_ip_prefix": "0.0.0.0/0",
            "protocol": protocol,
        }}
        if protocol == 'icmp':
            body['security_group_rule'].pop('port_range_min', None)
            body['security_group_rule'].pop('port_range_max', None)
        try:
            self.neutron.create_security_group_rule(body=body)
        except neutronclient.common.exceptions.Conflict as e:
            logger.error("error while creating a rule: %s" % e.message)
            pass

    def create_security_group(self, project_id, sec_g_name=None):
        if not sec_g_name:
            sec_g_name = sec_group_name
        sec_group = {}
        for sg in self.list_sec_group(project_id):
            if sg['name'] == sec_g_name:
                sec_group['security_group'] = sg
                break
        if len(sec_group) == 0:
            body = {"security_group": {
                'name': sec_g_name,
                'description': 'openbaton security group',
            }}
            sec_group = self.neutron.create_security_group(body=body)
            self.create_rule(sec_group, 'tcp')
            self.create_rule(sec_group, 'udp')
            self.create_rule(sec_group, 'icmp')
        self.sec_group = sec_group['security_group']
        return self.sec_group

    def list_sec_group(self, os_project_id):
        if not self.neutron:
            self.set_neutron(os_project_id)
        return [sec for sec in self.neutron.list_security_groups()['security_groups'] if
                (sec.get('tenant_id') is not None and sec.get('tenant_id') == os_project_id) or (
                    sec.get('project_id') is not None and sec.get('project_id') == os_project_id)]

    def get_vim_instance(self, tenant_name, username=None, password=None):
        if username:
            un = username
        else:
            un = self.username
        if password:
            pwd = password
        else:
            pwd = self.password

        logger.debug("Using tenant id: %s " % tenant_name)

        return {
            "name": "vim-instance-%s" % self.testbed_name,
            "authUrl": self.auth_url,
            "tenant": tenant_name,
            "username": un,
            "password": pwd,
            "securityGroups": [
                'default', sec_group_name
            ],
            "type": "openstack",
            "location": {
                "name": "Berlin",
                "latitude": "52.525876",
                "longitude": "13.314400"
            }
        }

    def list_images(self, tenant_id=None):
        if not self.nova:
            if not tenant_id:
                logger.error("Missing tenant_id!")
                raise OpenstackClientError('Missing tenant_id!')
            self.set_nova(tenant_id)
        try:
            imgs = self.nova.images.list()
            return imgs
        except:
            self.set_glance(tenant_id)
            return self.glance.images.list()

    def _get_tenant_id_from_name(self, tenant_name):
        if self.api_version == 2:
            tenants_list = self.keystone.tenants.list()
        else:
            tenants_list = self.keystone.projects.list()
        for tenant in tenants_list:
            if tenant.name == tenant_name:
                return tenant.id

    def set_glance(self, os_tenant_id):
        self.os_tenant_id = os_tenant_id
        self.glance = Glance('1', session=self._get_session(os_tenant_id))

    def _get_tenant_name_from_id(self, os_tenant_id):
        for t in self.list_tenants():
            if t.id == os_tenant_id:
                return t.name

    def create_user(self, username, password=None, tenant_id=None):
        for u in self.list_users():
            if hasattr(u, 'username'):
                u_username = u.username
            else:
                u_username = u.name
            if u_username == username:
                return u
        if not password:
            raise OpenstackClientError("Paswsord is needed to create user")
        if self.api_version == 2:
            return self.keystone.users.create(username, password, tenant_id=tenant_id)
        else:
            return self.keystone.users.create(name=username, password=password,
                                              project=self.get_project_from_id(tenant_id))

    def list_users(self):
        return self.keystone.users.list()

    def list_networks(self, project_id=None):
        if not self.neutron:
            if not project_id:
                raise OpenstackClientError("Missing project_id!")
            self.set_neutron(project_id)
        return self.neutron.list_networks(tenant_id=project_id)

    def list_subnets(self, project_id):
        if not self.neutron:
            if not project_id:
                raise OpenstackClientError("Missing project_id!")
            self.set_neutron(project_id)
        return self.neutron.list_subnets(tenant_id=project_id)

    def list_floatingips(self, project_id):
        if not self.neutron:
            if not project_id:
                raise OpenstackClientError("Missing project_id!")
            self.set_neutron(project_id)
        return self.neutron.list_floatingips(tenant_id=project_id)

    def list_routers(self, project_id):
        if not self.neutron:
            if not project_id:
                raise OpenstackClientError("Missing project_id!")
            self.set_neutron(project_id)
        return self.neutron.list_routers(tenant_id=project_id)

    def list_ports(self, project_id):
        if not self.neutron:
            if not project_id:
                raise OpenstackClientError("Missing project_id!")
            self.set_neutron(project_id)
        return self.neutron.list_ports(tenant_id=project_id)

    def list_keypairs(self, os_project_id=None):
        if not self.nova:
            if not os_project_id:
                raise OpenstackClientError("Missing project_id!")
            self.set_nova(os_project_id)
        return self.nova.keypairs.list()

    def list_domains(self):
        return self.keystone.domains.list()

    def get_project_from_id(self, tenant_id):
        for p in self.list_tenants():
            if p.id == tenant_id:
                return p
        raise OpenstackClientError("Project with id %s not found")

    def delete_user(self, username):
        try:
            self.keystone.users.delete(self.create_user(username=username))
        except:
            traceback.print_exc()
            logger.error("Not Able to delete user %s" % username)

    def delete_project(self, project_id):
        try:
            if self.api_version == 2:
                self.keystone.tenants.delete(project_id)
            else:
                self.keystone.projects.delete(project_id)
        except:
            traceback.print_exc()
            logger.error("Not Able to delete project %s" % project_id)

    def release_floating_ips(self, project_id):
        fips = self.list_floatingips(project_id).get('floatingips')
        for fip in fips:
            self.neutron.delete_floatingip(fip.get('id'))

    def delete_ports(self, project_id):
        ports = self.list_ports(project_id).get('ports')
        for port in ports:
            try:
                self.neutron.delete_port(port.get('id'))
            except Exception as e:
                pass

    def remove_gateway_routers(self, project_id):
        routers = self.list_routers(project_id).get('routers')
        for router in routers:
            self.neutron.remove_gateway_router(router.get('id'))

    def remove_interface_routers(self, project_id):
        routers = self.list_routers(project_id).get('routers')
        subnets = self.list_subnets(project_id).get('subnets')
        for router in routers:
            for subnet in subnets:
                body_value = {
                    'subnet_id': subnet.get('id'),
                }
                try:
                    self.neutron.remove_interface_router(router.get('id'), body_value)
                    break
                except Exception as e:
                    pass
            else:
                logger.warn('No subnet found that is associated to router {}'.format(router.get('id')))

    def delete_routers(self, project_id):
        routers = self.list_routers(project_id).get('routers')
        for router in routers:
            self.neutron.delete_router(router.get('id'))



    def delete_networks(self, project_id):
        networks = self.list_networks(project_id).get('networks')
        for nw in networks:
            self.neutron.delete_network(nw.get('id'))

    def delete_security_groups(self, project_id):
        sec_groups = self.list_sec_group(project_id)
        for sec_group in sec_groups:
            self.neutron.delete_security_group(sec_group.get('id'))




def _list_images_single_tenant(tenant_name, testbed, testbed_name):
    os_client = OSClient(testbed_name, testbed, tenant_name)
    result = []
    for image in os_client.list_images():
        logger.debug("%s" % image.name)
        result.append({
            'name': image.name,
            'testbed': testbed_name
        })
    return result


def list_images(tenant_name, testbed_name=None):
    openstack_credentials = get_openstack_credentials()
    images = []
    if not testbed_name:
        for name, testbed in openstack_credentials.items():
            logger.info("listing images for testbed %s" % name)
            try:
                images.extend(_list_images_single_tenant(tenant_name, testbed, name))
            except Exception as e:
                traceback.print_exc()
                logger.error("Error listing images for testbed: %s" % name)
                continue
    else:
        images = _list_images_single_tenant(tenant_name, openstack_credentials.get(testbed_name), testbed_name)
    return images


def create_os_project(username, password, tenant_name, testbed_name=None):
    openstack_credentials = get_openstack_credentials()
    os_tenants = {}
    if not testbed_name:
        for name, testbed in openstack_credentials.items():
            try:
                logger.info("Creating project on testbed: %s" % name)
                os_tenant_id, vim_instance = _create_single_project(tenant_name, testbed, name, username, password)
                logger.info("Created project %s on testbed: %s" % (os_tenant_id, name))
                os_tenants[name] = {'tenant_id': os_tenant_id, 'vim_instance': vim_instance}
            except:
                logger.error("Not able to create project in testbed %s" % name)
                traceback.print_exc()
                continue
    else:
        os_tenant_id, vim_instance = _create_single_project(tenant_name,
                                                            openstack_credentials[testbed_name],
                                                            testbed_name)
        os_tenants[testbed_name] = {'tenant_id': os_tenant_id, 'vim_instance': vim_instance}
    return os_tenants


def _create_single_project(tenant_name, testbed, testbed_name, username, password):
    os_client = OSClient(testbed_name, testbed)
    logger.info("Created OSClient for testbed %s" % testbed_name)
    admin_user = os_client.get_user()

    logger.debug("Got User %s" % admin_user)
    admin_role = os_client.get_role('admin')
    try:
        user_role = os_client.get_role('_member_')
    except:
        user_role = os_client.get_role('member')

    logger.debug("Got Role %s" % admin_role)
    for tenant in os_client.list_tenants():
        if tenant.name == tenant_name:
            logger.warn("Tenant with name or id %s exists already! I assume a double registration i will not do "
                        "anything :)" % tenant_name)
            logger.warn("returning tenant id %s" % tenant.id)

            exp_user = os_client.get_user(username)
            if not exp_user:
                exp_user = os_client.create_user(username, password)
                os_client.add_user_role(user=exp_user, role=user_role, tenant=tenant.id)
                os_client.add_user_role(user=admin_user, role=admin_role, tenant=tenant.id)
            if os_client.api_version == 2:
                vim_instance = os_client.get_vim_instance(tenant_name=tenant_name, username=username, password=password)
            else:
                vim_instance = os_client.get_vim_instance(tenant_name=tenant.id, username=username, password=password)
            return tenant.id, vim_instance

    tenant = os_client.create_tenant(tenant_name=tenant_name, description='softfire tenant for user %s' % tenant_name)
    logger.debug("Created tenant %s" % tenant)
    os_tenant_id = tenant.id
    logger.info("Created tenant with id: %s" % os_tenant_id)

    exp_user = os_client.create_user(username, password, os_tenant_id)
    os_client.add_user_role(user=admin_user, role=admin_role, tenant=os_tenant_id)
    os_client.add_user_role(user=exp_user, role=user_role, tenant=os_tenant_id)

    os_client = OSClient(testbed_name, testbed, project_id=os_tenant_id, tenant_name=tenant_name)

    try:
        ext_net = os_client.get_ext_net(testbed.get('ext_net_name'))

        if ext_net is None:
            logger.error(
                "A shared External Network called %s must exist! "
                "Please create one in your openstack instance" % testbed.get('ext_net_name')
            )
            raise OpenstackClientError("A shared External Network called softfire-network must exist! "
                                       "Please create one in your openstack instance")
        # networks, subnets, router_id = os_client.create_networks_and_subnets(ext_net)
        # logger.debug("Created Network %s, Subnet %s, Router %s" % (networks, subnets, router_id))

        fips = testbed.get("allocate-fip")
        if fips is not None and int(fips) > 0:
            try:
                os_client.allocate_floating_ips(int(fips), ext_net)
            except OpenstackClientError as e:
                logger.warn(e.args)

    except:
        logger.warning("Not able to get ext net")

    os_client.create_security_group(os_tenant_id)
    if os_client.api_version == 2:
        vim_instance = os_client.get_vim_instance(tenant_name=tenant_name, username=username, password=password)
    else:
        vim_instance = os_client.get_vim_instance(tenant_name=tenant.id, username=username, password=password)
    return os_tenant_id, vim_instance


def get_username_hash(username):
    return abs(hash(username))


def delete_tenant_and_user(username, testbed_tenants):
    openstack_credentials = get_openstack_credentials()
    for testbed_id, project_id in testbed_tenants.items():
        for testbed_name, credentials in openstack_credentials.items():
            if get_testbed_name_from_id(testbed_id) == testbed_name:
                os_client = OSClient(testbed_name, credentials)
                os_client.delete_security_groups(project_id)
                os_client.release_floating_ips(project_id)
                os_client.remove_gateway_routers(project_id)
                os_client.remove_interface_routers(project_id)
                os_client.delete_ports(project_id)
                os_client.delete_routers(project_id)
                os_client.delete_networks(project_id)
                os_client.delete_user(username)
                os_client.delete_project(project_id)


if __name__ == '__main__':
    for testbed_name, credentials in get_openstack_credentials().items():
        print("Executing list test on testbed %s" % testbed_name)
        client = OSClient(testbed_name, credentials)
        project_id = credentials.get("admin_project_id")
        print(client.list_images(project_id))
        print(client.list_tenants())
        print(client.list_users())
        print(client.list_roles())
        print(client.list_networks(project_id))
        print(client.list_keypairs(project_id))
        print(client.list_domains())
        print(client.list_sec_group(project_id))



import json

from org.openbaton.cli.agents.agents import OpenBatonAgentFactory
from org.openbaton.cli.errors.errors import NfvoException
from sdk.softfire.utils import get_config

from eu.softfire.nfv.utils.static_config import CONFIG_FILE_PATH
from eu.softfire.nfv.utils.utils import get_logger

logger = get_logger(__name__)


def get_vim_instance_test():
    return {
        "name":           "vim-instance-test",
        "authUrl":        'http://test.test.ts',
        "tenant":         'test',
        "username":       'test',
        "password":       'test',
        "securityGroups": [
            'default'
        ],
        "type":           "test",
        "location":       {
            "name":      "Test of nowhere",
            "latitude":  "5.525876",
            "longitude": "31.314400"
        }
    }


class OBClient(object):
    def __init__(self, project_name=None):
        """
        If username and password are passed, it will create a openbaton agent using these ones. if not it will use the
        configuration parameter ones.

        :param project_name: the project of a specific user.
        """
        https = get_config("nfvo", "https", CONFIG_FILE_PATH).lower() == 'true'

        username = get_config("nfvo", "username", CONFIG_FILE_PATH)

        password = get_config("nfvo", "password", CONFIG_FILE_PATH)

        nfvo_ip = get_config("nfvo", "ip", CONFIG_FILE_PATH)
        nfvo_port = get_config("nfvo", "port", CONFIG_FILE_PATH)
        self.agent_factory = OpenBatonAgentFactory(nfvo_ip=nfvo_ip,
                                                   nfvo_port=nfvo_port,
                                                   https=https,
                                                   version=1,
                                                   username=username,
                                                   password=password,
                                                   project_id=None)
        if project_name:
            self.project_id = self._get_project_id(project_name)
            if self.project_id is None and project_name is not None:
                logger.warning('Project ID is None. No project found with the name {}'.format(project_name))

    def _get_project_id(self, project_name):
        project_agent = self.agent_factory.get_project_agent()
        for project in json.loads(project_agent.find()):
            if project.get('name') == project_name:
                return project.get('id')
        return None

    def list_nsds(self):
        return json.loads(self.agent_factory.get_ns_descriptor_agent(self.project_id).find())

    def create_nsr(self, nsd_id, body=None):
        return self.agent_factory.get_ns_records_agent(self.project_id).create(nsd_id, body)

    def delete_nsr(self, nsr_id):
        return self.agent_factory.get_ns_records_agent(self.project_id).delete(nsr_id)

    def create_project(self, project):
        for p in json.loads(self.list_projects()):
            if p.get('name') == project.get('name'):
                return p
        if isinstance(project, dict):
            project = json.dumps(project)
        ob_project = self.agent_factory.get_project_agent().create(project)
        self.project_id = ob_project.get('id')
        return ob_project

    def create_user(self, user):

        for us in json.loads(self.list_users()):
            if us.get('username') == user.get('username'):
                return us

        if isinstance(user, dict):
            user = json.dumps(user)
        return self.agent_factory.get_user_agent(self.project_id).create(user)

    def create_vim_instance(self, vim_instance):
        for vi in self.list_vim_instances():
            if vi.get('name') == vim_instance.get('name'):
                return vi
        if isinstance(vim_instance, dict):
            vim_instance = json.dumps(vim_instance)

        logger.debug("Posting vim %s" % vim_instance)
        return self.agent_factory.get_vim_instance_agent(self.project_id).create(vim_instance)

    def list_users(self):
        return self.agent_factory.get_user_agent(self.project_id).find()

    def list_projects(self):
        return self.agent_factory.get_project_agent().find()

    def list_vim_instances(self):
        return json.loads(self.agent_factory.get_vim_instance_agent(self.project_id).find())

    def list_images_network_flavors(self):
        images = []
        networks = []
        flavors = []
        vim_instance_agent = self.agent_factory.get_vim_instance_agent(self.project_id)

        for vim_instance in json.loads(vim_instance_agent.find()):

            vim_instance_name = vim_instance.get("name")
            if "-" in vim_instance_name:
                testbed_name = vim_instance_name.split("-")[-1]
            else:
                testbed_name = vim_instance_name
            for img in vim_instance.get("images"):
                images.append(
                    {
                        "testbed": testbed_name,
                        "name": img.get("name")
                    }
                )
            for net in vim_instance.get("networks"):
                networks.append(
                    {
                        "testbed": testbed_name,
                        "name": net.get("name")
                    }
                )
            for flavor in vim_instance.get("flavours"):
                flavors.append(
                    {
                        "testbed": testbed_name,
                        "name": flavor.get("flavour_key")
                    }
                )
        return images, networks, flavors

    def upload_package(self, package_path, name=None):
        package_agent = self.agent_factory.get_vnf_package_agent(self.project_id)
        try:
            return package_agent.create(package_path)
        except NfvoException as e:
            if not name:
                raise e
            for nsd in json.loads(self.agent_factory.get_ns_descriptor_agent(self.project_id).find()):
                for vnfd in nsd.get('vnfd'):
                    if vnfd.get('name') == name:
                        return {"id": vnfd.get('id')}
            raise e

    def create_nsd(self, nsd):
        if isinstance(nsd, dict):
            nsd = json.dumps(nsd)

        logger.debug("Uplading really: \n%s" % nsd)
        return self.agent_factory.get_ns_descriptor_agent(self.project_id).create(nsd)

    def get_nsd(self, nsd_id):
        return self.agent_factory.get_ns_descriptor_agent(self.project_id).find(nsd_id)

    def delete_nsd(self, nsd_id):
        self.agent_factory.get_ns_descriptor_agent(self.project_id).delete(nsd_id)

    def delete_vnfd(self, vnfd_id):
        self.agent_factory.get_vnf_descriptor_agent(self.project_id).delete(vnfd_id)

    def get_nsr(self, nsr_id):
        return self.agent_factory.get_ns_records_agent(self.project_id).find(nsr_id)

    def import_key(self, ssh_pub_key, name):

        key_agent = self.agent_factory.get_key_agent(self.project_id)
        for key in json.loads(key_agent.find()):
            if key.get('name') == name:
                key_agent.delete(key.get('id'))
                break

        key_agent.create(
            json.dumps(
                {
                    'name': name,
                    'projectId': self.project_id,
                    'publicKey': ssh_pub_key
                }
            )
        )

    def create_nsd_from_csar(self, location):
        return self.agent_factory.get_csarnsd_agent(self.project_id).create(location)

    def delete_user(self, username):
        for u in json.loads(self.list_users()):
            if u.get('username') == username:
                self.agent_factory.get_user_agent(self.project_id).delete(u.get('id'))

    def delete_project(self, ob_project_id):
        self.agent_factory.get_project_agent().delete(ob_project_id)

    def list_nsrs(self):
        return json.loads(self.agent_factory.get_ns_records_agent(self.project_id).find())

    def delete_vim_instance(self, _vim_id):
        self.agent_factory.get_vim_instance_agent(self.project_id).delete(_vim_id)



from sdk.softfire.main import start_manager

from eu.softfire.nfv.core.NfvManager import NfvManager, UpdateStatusThread


def start():
    nfv_manager = NfvManager('/etc/softfire/nfv-manager.ini')
    thread = UpdateStatusThread(nfv_manager)
    thread.start()
    try:
        start_manager(nfv_manager)
    finally:
        thread.stop()
        thread.join()

if __name__ == '__main__':
    start()
